@@ -5,6 +5,7 @@ sca2d to analyse the scad code.
 
 import os
 from copy import copy
+import re
 from sca2d.scope import ScopeContents
 from sca2d import utilities
 from sca2d.scadclasses import (UseIncStatment,
@@ -15,7 +16,7 @@ from sca2d.scadclasses import (UseIncStatment,
                                FunctionDef)
 from sca2d.messages import Message
 from sca2d.definitions import SCAD_VARS, SCAD_MODS, SCAD_FUNCS
-
+from sca2d.patterns import GOBAL_PATTERN
 
 class OuterScope(ScopeContents):
     """
@@ -30,10 +31,13 @@ class OuterScope(ScopeContents):
     on all internal scopes.
     """
 
-    def __init__(self, tree, filename):
+    def __init__(self, tree, scad_code, filename):
         self._filename = filename
         self._used_files = []
         self._included_files = []
+        self._scad_code = scad_code
+        #process the scad code into self._file_contents (a list each line)
+        self._process_scad_code()
         # This will begin parsing the tree for this file and all sub trees
         # See ScopeContents._parse_scope
         super().__init__(tree, None, top_level=True)
@@ -49,6 +53,13 @@ class OuterScope(ScopeContents):
         '''
         return self._filename
 
+    @property
+    def file_contents(self):
+        '''
+        Returns the contents of the scad file.
+        '''
+        return self._file_contents
+
     @property
     def used_files(self):
         '''
@@ -63,6 +74,35 @@ class OuterScope(ScopeContents):
         '''
         return copy(self._included_files)
 
+    def get_line_and_column(self, char_index):
+        """
+        Return the line and column number for the python index of a given
+        character
+        """
+        line = next(i for i, line_end in enumerate(self._line_endings) if  line_end > char_index) + 1
+        column = char_index - self._line_endings[line-2]
+        return line, column
+
+    def _process_scad_code(self):
+        if self._scad_code is None:
+            self._file_contents = []
+            self._line_endings = []
+        else:
+            self._line_endings = [i for i, char in enumerate(self._scad_code) if char == "\n"]
+            # Python has already ensured that \n is used for newlines
+            # on all platforms
+            self._file_contents = self._scad_code.split('\n')
+
+    def _check_trailing_whitespace(self):
+        #matches are either comments or trailing white spaces in code
+        #The first group is the trailing whitespace. This is empty for comments
+        matches = re.finditer(r'/\*(?:.|\n)*?\*/|//.*|( $)', self._scad_code, re.MULTILINE)
+        # Get only whitepace matches
+        matches = [match_obj for match_obj in matches if match_obj.group(1) is not None]
+        for match_obj in matches:
+            line, col = self.get_line_and_column(match_obj.start())
+            self._record_message("I0006", DummyTree(data='whitespace', line=line, column=col+1))
+
     def _printed_var_lists(self, indent=2, this_indent=0):
         def print_list(plist):
             return '[' + ', '.join([str(item) for item in plist]) + ']'
@@ -90,8 +130,21 @@ class OuterScope(ScopeContents):
         """
         self._check_pointless_termination()
         self._check_pointless_scope()
-        self._check_complicated_argument()
+        self._check_complexity()
+        self._check_invalid_attributes()
         self._check_defintions(analyser)
+        self._check_globals()
+        self._check_trailing_whitespace()
+
+    def _check_globals(self):
+        global_definitions = [var for var in self._assigned_vars if not var.is_special()]
+        n_globals = len(global_definitions)
+        if n_globals>3:
+            self._record_message("I1002", self.tree, [n_globals])
+        for var in global_definitions:
+            global_match = re.match(GOBAL_PATTERN, var.name)
+            if global_match is None:
+                self._record_message("I3001", var.tree, [var.name])
 
     def _check_defintions(self, analyser):
         defs_and_msgs = BUILT_IN_SCOPE.get_external_defintions(analyser)
@@ -102,8 +155,8 @@ class OuterScope(ScopeContents):
 
         #To get the definitions from use and included files we run the same import
         #as any other file including this file with an include statment. We just
-        # do not count the definitions from this file. Hence count_own being off.
-        defs_and_msgs = self.get_external_defintions(analyser, count_own=False)
+        # do not count the definitions from this file.
+        defs_and_msgs = self.get_external_defintions(analyser, root_file=True)
         all_def_var += defs_and_msgs[0]
         all_def_mod += defs_and_msgs[1]
         all_def_func += defs_and_msgs[2]
@@ -112,12 +165,49 @@ class OuterScope(ScopeContents):
             self.messages.append(message)
 
         self._check_duplicate_external_defs(all_def_var, all_def_mod, all_def_func)
-        self.propogate_defs_and_use(all_def_var, all_def_mod, all_def_func)
+        all_uses = self.propogate_defs_and_use(all_def_var, all_def_mod, all_def_func)
+        self._check_if_external_files_needed(analyser, defs_and_msgs[:3], all_uses)
+
+    def _check_if_external_files_needed(self, analyser, all_defs, all_uses):
+
+        for used_file in self._used_files:
+            used_scope = analyser.get_scope_from_file(used_file)
+            if not self._is_file_needed(used_scope, all_defs, all_uses, file_included=False):
+                self._record_message("W1001", used_file.tree, [used_file.filename])
+        for included_file in self._included_files:
+            included_scope = analyser.get_scope_from_file(included_file)
+            if not self._is_file_needed(included_scope, all_defs, all_uses, file_included=True):
+                self._record_message("W1002", included_file.tree, [included_file.filename])
+
+    def _is_file_needed(self, file_scope, all_defs, all_uses, file_included):
+        all_def_var, all_def_mod, all_def_func = all_defs
+        all_var_use, all_mod_use, all_func_use = all_uses
+
+        file_needed = False
+
+        for mod in all_def_mod:
+            if mod.included_by is file_scope:
+                if mod in all_mod_use:
+                    file_needed = True
+                    continue
+        if not file_needed:
+            for func in all_def_func:
+                if func.included_by is file_scope:
+                    if func in all_func_use:
+                        file_needed = True
+                        continue
+        if file_included and not file_needed:
+            for var in all_def_var:
+                if var.included_by is file_scope:
+                    if var in all_var_use:
+                        file_needed = True
+                        continue
+        return file_needed
 
     def get_external_defintions(self,
                                 analyser,
                                 breadcrumbs=None,
-                                count_own=True,
+                                root_file=False,
                                 inc_only=False):
         """
         The definitions passed onto a file that when this file is included with and
@@ -133,7 +223,7 @@ class OuterScope(ScopeContents):
         all_def_func = []
         #The message raised when getting external definitions
         all_msgs = []
-        if count_own:
+        if not root_file:
             all_def_var += self._assigned_vars
             all_def_mod += self._defined_modules
             all_def_func += self._defined_functions
@@ -156,6 +246,10 @@ class OuterScope(ScopeContents):
         all_def_func += defs[2]
         all_msgs += defs[3]
 
+        if not root_file:
+            for item in all_def_mod+all_def_func+all_def_var:
+                item.included_by=self
+
         return all_def_var, all_def_mod, all_def_func, all_msgs
 
     def _check_and_append_breadcrumbs(self, breadcrumbs):
@@ -189,13 +283,36 @@ class OuterScope(ScopeContents):
         for tree in subtrees:
             self._record_message("I0002", tree)
 
+    def _check_complexity(self):
+        self._check_complicated_argument()
+        self._check_complicated_assignment()
+
     def _check_complicated_argument(self):
-        subtrees = utilities.get_all_matching_subtrees(self._tree, 'arg')
-        for tree in subtrees:
-            tokens = utilities.get_all_tokens(tree)
-            n_tokens = len(tokens)
-            if n_tokens>=8:
-                self._record_message("I1001", tree, [n_tokens])
+        #only check modules as functions are checked by _check_complicated_assignment
+        mod_head_trees = utilities.get_all_matching_subtrees(self._tree, 'module_header')
+        for mod_head_tree in mod_head_trees:
+            arg_trees = utilities.get_all_matching_subtrees(mod_head_tree, 'arg')
+            for arg_tree in arg_trees:
+                complexity = utilities.estimate_complexity(arg_tree)
+                if complexity>=8:
+                    self._record_message("I1001", arg_tree, [complexity])
+
+    def _check_complicated_assignment(self):
+        var_assign_trees = utilities.get_all_matching_subtrees(self._tree,
+                                                               'variable_assignment',
+                                                               include_nested=False)
+        var_assign_trees += utilities.get_all_matching_subtrees(self._tree,
+                                                                'control_assignment',
+                                                                include_nested=False)
+        func_scope_trees = utilities.get_all_matching_subtrees(self._tree,
+                                                               'function_scope',
+                                                               include_nested=False)
+        expr_trees = [tree.children[1] for tree in var_assign_trees]
+        expr_trees += [tree.children[0] for tree in func_scope_trees]
+        for tree in expr_trees:
+            complexity = utilities.estimate_complexity(tree)
+            if complexity>=10:
+                self._record_message("I1002", tree, [complexity])
 
     def _check_invalid_attributes(self):
         attrs = utilities.get_all_matching_tokens(self._tree, "ATTRIBUTE")
@@ -253,9 +370,9 @@ class NonFileOuterScope(OuterScope):
     directory.
     """
 
-    def __init__(self, tree):
+    def __init__(self, tree, scad_code):
         filename = os.path.join(os.getcwd(), 'INPUT_CODE')
-        super().__init__(tree, filename)
+        super().__init__(tree, scad_code, filename)
 
 class BuiltInScope(OuterScope):
     """
@@ -263,7 +380,7 @@ class BuiltInScope(OuterScope):
     """
 
     def __init__(self):
-        super().__init__(DummyTree(), 'BUILT_IN')
+        super().__init__(DummyTree(), None, 'BUILT_IN')
         self._set_built_in_vars()
         self._set_built_in_mods()
         self._set_built_in_funcs()
